/*
    Dump utility for Agere/Lucent, Symbol and Intersil/Conexant 802.11b
    cards.  It dumps all RIDs (resources) in human readable format.

    Copyright (C) 2004  Pavel Roskin <proski@gnu.org>
    Copyright (C) 2004  Holger Schurig

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define PROGRAM_VERSION "0.2"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/wireless.h>

#ifndef SIOCIWFIRSTPRIV
#define SIOCIWFIRSTPRIV	0x8BE0
#endif

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

#define read_le16(x) ((x)[0] + ((x)[1] << 8))

#undef min_t
#define min_t(type,x,y) \
	({ type __x = (x); type __y = (y); __x < __y ? __x: __y; })

#define HERMES_RID_NICID 0xfd0b

#define MAX_RID_LEN 1024

typedef enum {
	FMT_WORDS,
	FMT_BYTES,
	FMT_STRING,
	FMT_BSTRING,
	FMT_XSTRING,
	FMT_FWSTRING,
	FMT_CIS,
	FMT_UNKNOWN
} display_t;

typedef enum {
	DRIVER_ORINOCO,
	DRIVER_HOSTAP,
	DRIVER_UNKNOWN
} driver_t;

#define FW_AGERE		1
#define FW_INTERSIL		2
#define FW_SYMBOL		4

#define FW_INTERSIL_SYMBOL	(FW_INTERSIL | FW_SYMBOL)
#define FW_COMMON		(FW_AGERE | FW_INTERSIL | FW_SYMBOL)

/* Global variables */
static int sock;
static char *iface;
static display_t out_type = FMT_UNKNOWN;
static driver_t driver_type = DRIVER_UNKNOWN;
static int fw_type;

struct record_item {
	u16 rid;
	char *name;
	int displaytype;
	int fwtypes;
};

static struct record_item record_table[] = {
	{0xfc00, "PortType", FMT_WORDS, FW_COMMON},
	{0xfc01, "OwnMACAddr", FMT_BYTES, FW_COMMON},
	{0xfc02, "DesiredSSID", FMT_STRING, FW_COMMON},
	{0xfc03, "OwnChannel", FMT_WORDS, FW_COMMON},
	{0xfc04, "OwnSSID", FMT_STRING, FW_COMMON},
	{0xfc05, "OwnATIMWindow", FMT_WORDS, FW_COMMON},
	{0xfc06, "SystemScale", FMT_WORDS, FW_COMMON},
	{0xfc07, "MaxDataLength", FMT_WORDS, FW_COMMON},
	{0xfc08, "WDSAddress", FMT_BYTES, FW_COMMON},
	{0xfc09, "PMEnabled", FMT_WORDS, FW_COMMON},
	{0xfc0a, "PMEPS", FMT_WORDS, FW_COMMON},
	{0xfc0b, "MulticastReceive", FMT_WORDS, FW_COMMON},
	{0xfc0c, "MaxSleepDuration", FMT_WORDS, FW_COMMON},
	{0xfc0d, "PMHoldoverDuration", FMT_WORDS, FW_COMMON},
	{0xfc0e, "OwnName", FMT_STRING, FW_COMMON},
	{0xfc0f, "MKKFlag", FMT_WORDS, FW_SYMBOL},
	{0xfc10, "OwnDTIMPeriod", FMT_WORDS, FW_COMMON},
	{0xfc11, "WDSAddress1", FMT_BYTES, FW_COMMON},
	{0xfc12, "WDSAddress2", FMT_BYTES, FW_COMMON},
	{0xfc13, "WDSAddress3", FMT_BYTES, FW_COMMON},
	{0xfc14, "WDSAddress4", FMT_BYTES, FW_COMMON},
	{0xfc15, "WDSAddress5", FMT_BYTES, FW_COMMON},
	{0xfc16, "WDSAddress6", FMT_BYTES, FW_COMMON},
	{0xfc17, "MulticastPMBuffering", FMT_WORDS, FW_COMMON},
	{0xfc18, "RejectAny", FMT_WORDS, FW_AGERE},
	{0xfc20, "Encryption", FMT_WORDS, FW_AGERE},
	{0xfc20, "PreferredBSSID", FMT_BYTES, FW_SYMBOL},
	{0xfc21, "Authentication", FMT_WORDS, FW_AGERE},
	{0xfc21, "MandatoryBSSID", FMT_BYTES, FW_SYMBOL},
	{0xfc22, "ExclUnencrypted", FMT_WORDS, FW_AGERE},
	{0xfc22, "Authentication0", FMT_WORDS, FW_SYMBOL},
	{0xfc23, "MulticastRate", FMT_WORDS, FW_AGERE},
	{0xfc23, "WEPDefaultKeyID", FMT_WORDS, FW_INTERSIL_SYMBOL},
	{0xfc24, "IntraBSSDelay", FMT_WORDS, FW_AGERE},
	{0xfc24, "DefaultKey0", FMT_BYTES, FW_INTERSIL_SYMBOL},
	{0xfc25, "MWORobust", FMT_WORDS, FW_AGERE},
	{0xfc25, "DefaultKey1", FMT_BYTES, FW_INTERSIL_SYMBOL},
	{0xfc26, "LoadBalancing", FMT_WORDS, FW_AGERE},
	{0xfc26, "DefaultKey2", FMT_BYTES, FW_INTERSIL_SYMBOL},
	{0xfc27, "MediumDistribution", FMT_WORDS, FW_AGERE},
	{0xfc27, "DefaultKey3", FMT_BYTES, FW_INTERSIL_SYMBOL},
	{0xfc28, "RxAllGroupFilter", FMT_WORDS, FW_AGERE},
	{0xfc28, "WEPFlags", FMT_WORDS, FW_INTERSIL_SYMBOL},
	{0xfc29, "CountryInfo", FMT_STRING, FW_AGERE},
	{0xfc29, "WEPKeyMapping", FMT_BYTES, FW_INTERSIL},
	{0xfc2a, "TxPower", FMT_WORDS, FW_AGERE},
	{0xfc2a, "Authentication", FMT_WORDS, FW_INTERSIL_SYMBOL},
	{0xfc2b, "MaxAssociatedStations", FMT_WORDS, FW_INTERSIL},
	{0xfc2b, "KeyLength", FMT_WORDS, FW_SYMBOL},
	{0xfc2c, "TxControl", FMT_WORDS, FW_INTERSIL},
	{0xfc2d, "RoamingMode", FMT_WORDS, FW_INTERSIL},
	{0xfc2e, "HostAuthentication", FMT_WORDS, FW_INTERSIL},
	{0xfc2f, "EnableKerberos", FMT_WORDS, FW_SYMBOL},
	{0xfc30, "ConnectionControl", FMT_WORDS, FW_AGERE},
	{0xfc30, "RcvCrcError", FMT_WORDS, FW_INTERSIL},
	{0xfc30, "AddSessionKey", FMT_WORDS, FW_SYMBOL},
	{0xfc31, "MMLife", FMT_WORDS, FW_INTERSIL},
	{0xfc31, "DelSessionKey", FMT_WORDS, FW_SYMBOL},
	{0xfc32, "AltRetryCount", FMT_WORDS, FW_INTERSIL},
	{0xfc33, "BeaconInterval", FMT_WORDS, FW_INTERSIL},
	{0xfc34, "APPCFInfo", FMT_WORDS, FW_INTERSIL},
	{0xfc35, "STAPCFInfo", FMT_WORDS, FW_INTERSIL},
	{0xfc37, "PriorityQUsase", FMT_WORDS, FW_INTERSIL},
	{0xfc40, "TIMCtrl", FMT_WORDS, FW_INTERSIL},
	{0xfc42, "Thirty2Tally", FMT_WORDS, FW_INTERSIL},
	{0xfc43, "EnhSecurity", FMT_WORDS, FW_INTERSIL},
	{0xfc46, "dBmAdjust", FMT_WORDS, FW_INTERSIL},
	{0xfc48, "GenericElement", FMT_WORDS, FW_INTERSIL},
	{0xfc49, "PropagationDelay", FMT_WORDS, FW_INTERSIL},
	{0xfc80, "GroupAddress", FMT_BYTES, FW_COMMON},
	{0xfc81, "CreateIBSS", FMT_WORDS, FW_COMMON},
	{0xfc82, "FragmenationThreshold", FMT_WORDS, FW_COMMON},
	{0xfc83, "RTSThreshold", FMT_WORDS, FW_COMMON},
	{0xfc84, "TxRateControl", FMT_WORDS, FW_COMMON},
	{0xfc85, "PromiscuousMode", FMT_WORDS, FW_COMMON},
	{0xfc86, "WakeOnLAN", FMT_WORDS, FW_AGERE},
	{0xfc87, "WakeOnLANPatterns", FMT_BYTES, FW_AGERE},
	{0xfc87, "AntennaDiversity", FMT_WORDS, FW_SYMBOL},
	{0xfc88, "Options", FMT_WORDS, FW_SYMBOL},
	{0xfc89, "IFFrequency", FMT_WORDS, FW_SYMBOL},
	{0xfc8a, "BasicRates", FMT_WORDS, FW_SYMBOL},
	{0xfc8c, "Preamble", FMT_WORDS, FW_SYMBOL},
	{0xfc8e, "TxPower", FMT_WORDS, FW_SYMBOL},
	{0xfc90, "FragmentationThreshold0", FMT_WORDS, FW_COMMON},
	{0xfc91, "FragmentationThreshold1", FMT_WORDS, FW_COMMON},
	{0xfc92, "FragmentationThreshold2", FMT_WORDS, FW_COMMON},
	{0xfc93, "FragmentationThreshold3", FMT_WORDS, FW_COMMON},
	{0xfc94, "FragmentationThreshold4", FMT_WORDS, FW_COMMON},
	{0xfc95, "FragmentationThreshold5", FMT_WORDS, FW_COMMON},
	{0xfc96, "FragmentationThreshold6", FMT_WORDS, FW_COMMON},
	{0xfc97, "RTSThreshold0", FMT_WORDS, FW_COMMON},
	{0xfc98, "RTSThreshold1", FMT_WORDS, FW_COMMON},
	{0xfc99, "RTSThreshold2", FMT_WORDS, FW_COMMON},
	{0xfc9a, "RTSThreshold3", FMT_WORDS, FW_COMMON},
	{0xfc9b, "RTSThreshold4", FMT_WORDS, FW_COMMON},
	{0xfc9c, "RTSThreshold5", FMT_WORDS, FW_COMMON},
	{0xfc9d, "RTSThreshold6", FMT_WORDS, FW_COMMON},
	{0xfc9e, "TxRateControl0", FMT_WORDS, FW_COMMON},
	{0xfc9f, "TxRateControl1", FMT_WORDS, FW_COMMON},
	{0xfca0, "TxRateControl2", FMT_WORDS, FW_COMMON},
	{0xfca1, "TxRateControl3", FMT_WORDS, FW_COMMON},
	{0xfca2, "TxRateControl4", FMT_WORDS, FW_COMMON},
	{0xfca3, "TxRateControl5", FMT_WORDS, FW_COMMON},
	{0xfca4, "TxRateControl6", FMT_WORDS, FW_COMMON},
	{0xfca5, "TxInhibit", FMT_WORDS, FW_SYMBOL},
	{0xfca6, "TxClearStats", FMT_WORDS, FW_SYMBOL},
	{0xfcab, "HostScan", FMT_WORDS, FW_SYMBOL},
	{0xfcad, "ACKTimeout", FMT_WORDS, FW_SYMBOL},
	{0xfcb0, "DefaultKeys", FMT_BYTES, FW_AGERE},
	{0xfcb0, "ShortPreamble", FMT_WORDS, FW_INTERSIL},
	{0xfcb1, "TxKeyID", FMT_WORDS, FW_AGERE},
	{0xfcb1, "ExcludeLongPreamble", FMT_WORDS, FW_INTERSIL},
	{0xfcb2, "ScanSSID", FMT_STRING, FW_AGERE},
	{0xfcb2, "AuthResponceTimeout", FMT_WORDS, FW_INTERSIL},
	{0xfcb3, "BasicRates", FMT_WORDS, FW_INTERSIL},
	{0xfcb4, "AddTKIPDefaultKey", FMT_WORDS, FW_AGERE},
	{0xfcb4, "SupportedRates", FMT_WORDS, FW_INTERSIL},
	{0xfcb5, "SetWPAKeyMgmtSuite", FMT_WORDS, FW_AGERE},
	{0xfcb5, "FallbackCtrl", FMT_WORDS, FW_INTERSIL},
	{0xfcb6, "RemoveTKIPDefaultKey", FMT_WORDS, FW_AGERE},
	{0xfcb6, "WEPKeyStatus", FMT_WORDS, FW_INTERSIL},
	{0xfcb7, "AddTKIPMappedKey", FMT_WORDS, FW_AGERE},
	{0xfcb7, "WEPKeyMapIndex", FMT_WORDS, FW_INTERSIL},
	{0xfcb8, "RemoveTKIPMappedKey", FMT_WORDS, FW_AGERE},
	{0xfcb8, "BroadcastKeyID", FMT_WORDS, FW_INTERSIL},
	{0xfcb9, "SetWPACapabilitiesInfo", FMT_WORDS, FW_AGERE},
	{0xfcb9, "EntSecFlag", FMT_WORDS, FW_INTERSIL},
	{0xfcba, "CachedPMKAddress", FMT_BYTES, FW_AGERE},
	{0xfcba, "PassiveScanCtrl", FMT_WORDS, FW_INTERSIL},
	{0xfcbb, "RemoveCachedPMKAddress", FMT_WORDS, FW_AGERE},
	{0xfcbb, "SSNHandlingMode", FMT_WORDS, FW_INTERSIL},
	{0xfcbc, "MDCControl", FMT_WORDS, FW_INTERSIL},
	{0xfcbd, "MDCCountry", FMT_XSTRING, FW_INTERSIL},
	{0xfcbe, "TxPowerMax", FMT_WORDS, FW_INTERSIL},
	{0xfcbf, "LFOEnabled", FMT_WORDS, FW_INTERSIL},
	{0xfcc0, "HandoverAddress", FMT_BYTES, FW_AGERE},
	{0xfcc0, "CapInfo", FMT_WORDS, FW_INTERSIL},
	{0xfcc1, "ListenInterval", FMT_WORDS, FW_INTERSIL},
	{0xfcc2, "ScanChannel", FMT_WORDS, FW_AGERE},
	{0xfcc2, "AntennaDiversity", FMT_WORDS, FW_INTERSIL},
	{0xfcc4, "DisassociateAddress", FMT_BYTES, FW_AGERE},
	{0xfcc4, "LedCtrl", FMT_WORDS, FW_INTERSIL},
	{0xfcc5, "ProbeDataRate", FMT_WORDS, FW_AGERE},
	{0xfcc5, "HFODelay", FMT_WORDS, FW_INTERSIL},
	{0xfcc6, "DisallowedBSSID", FMT_BYTES, FW_INTERSIL},
	{0xfce0, "TickTime", FMT_WORDS, FW_COMMON},
	{0xfce1, "ScanRequest", FMT_WORDS, FW_INTERSIL},
	{0xfce2, "JoinRequest", FMT_BYTES, FW_INTERSIL},
	{0xfce3, "AuthenticateStation", FMT_WORDS, FW_INTERSIL},
	{0xfce4, "ChannelInfoRequest", FMT_WORDS, FW_INTERSIL},
	{0xfce5, "HostScan", FMT_WORDS, FW_INTERSIL},
	{0xfce6, "AssociateStation", FMT_BYTES, FW_INTERSIL},
	{0xfd00, "MaxLoadTime", FMT_WORDS, FW_COMMON},
	{0xfd01, "DownloadBuffer", FMT_WORDS, FW_COMMON},
	{0xfd02, "PRIIdentity", FMT_WORDS, FW_COMMON},
	{0xfd03, "PRISupRange", FMT_WORDS, FW_COMMON},
	{0xfd04, "CFIActRangesPRI", FMT_WORDS, FW_COMMON},
	{0xfd05, "PrimaryVersion", FMT_XSTRING, FW_SYMBOL},
	{0xfd06, "PrimaryDate", FMT_XSTRING, FW_SYMBOL},
	{0xfd09, "NICHSISupRange", FMT_WORDS, FW_AGERE},
	{0xfd0a, "NICSerialNumber", FMT_XSTRING, FW_COMMON},
	{0xfd0b, "NICIdentity", FMT_WORDS, FW_COMMON},
	{0xfd0c, "MFISupRange", FMT_WORDS, FW_COMMON},
	{0xfd0d, "CFISupRange", FMT_WORDS, FW_COMMON},
	{0xfd10, "ChannelList", FMT_WORDS, FW_COMMON},
	{0xfd11, "RegulatoryDomains", FMT_BSTRING, FW_COMMON},
	{0xfd12, "TempType", FMT_WORDS, FW_COMMON},
	{0xfd13, "CIS", FMT_CIS, FW_COMMON},
	{0xfd14, "CardProfile", FMT_WORDS, FW_AGERE},
	{0xfd14, "CountryCode", FMT_WORDS, FW_SYMBOL},
	{0xfd15, "CountryName", FMT_XSTRING, FW_SYMBOL},
	{0xfd16, "MKKCallSign", FMT_WORDS, FW_SYMBOL},
	{0xfd17, "MKKRSSI", FMT_WORDS, FW_SYMBOL},
	{0xfd18, "Ordinals", FMT_WORDS, FW_SYMBOL},
	{0xfd20, "STAIdentity", FMT_WORDS, FW_COMMON},
	{0xfd21, "STASupRange", FMT_WORDS, FW_COMMON},
	{0xfd22, "MFIActRanges", FMT_WORDS, FW_COMMON},
	{0xfd23, "CFIActRangesSTA", FMT_WORDS, FW_COMMON},
	{0xfd24, "NICBusType", FMT_WORDS, FW_AGERE},
	{0xfd24, "ProductName", FMT_STRING, FW_INTERSIL},
	{0xfd24, "SecondaryVersion", FMT_XSTRING, FW_SYMBOL},
	{0xfd25, "SecondaryDate", FMT_XSTRING, FW_SYMBOL},
	{0xfd40, "PortStatus", FMT_WORDS, FW_COMMON},
	{0xfd41, "CurrentSSID", FMT_STRING, FW_COMMON},
	{0xfd42, "CurrentBSSID", FMT_BYTES, FW_COMMON},
	{0xfd43, "CommsQuality", FMT_WORDS, FW_COMMON},
	{0xfd44, "CurrentTxRate", FMT_WORDS, FW_COMMON},
	{0xfd45, "CurrentBeaconInterval", FMT_WORDS, FW_COMMON},
	{0xfd46, "CurrentScaleThresholds", FMT_WORDS, FW_COMMON},
	{0xfd47, "ProtocolRspTime", FMT_WORDS, FW_COMMON},
	{0xfd48, "ShortRetryLimit", FMT_WORDS, FW_COMMON},
	{0xfd49, "LongRetryLimit", FMT_WORDS, FW_COMMON},
	{0xfd4a, "MaxTransmitLifetime", FMT_WORDS, FW_COMMON},
	{0xfd4b, "MaxReceiveLifetime", FMT_WORDS, FW_COMMON},
	{0xfd4c, "CFPollable", FMT_WORDS, FW_COMMON},
	{0xfd4d, "AuthenticationAlgorithms", FMT_WORDS, FW_COMMON},
	{0xfd4e, "AuthenticationType", FMT_WORDS, FW_AGERE},
	{0xfd4f, "PrivacyOptionImplemented", FMT_WORDS, FW_COMMON},
	{0xfd50, "CurrentRemoteRates", FMT_WORDS, FW_AGERE},
	{0xfd51, "CurrentUsedRates", FMT_WORDS, FW_AGERE},
	{0xfd51, "dBmCommsQuality", FMT_WORDS, FW_INTERSIL},
	{0xfd52, "CurrentSystemScale", FMT_WORDS, FW_AGERE},
	{0xfd80, "CurrentTxRate1", FMT_WORDS, FW_COMMON},
	{0xfd81, "CurrentTxRate2", FMT_WORDS, FW_COMMON},
	{0xfd82, "CurrentTxRate3", FMT_WORDS, FW_COMMON},
	{0xfd83, "CurrentTxRate4", FMT_WORDS, FW_COMMON},
	{0xfd84, "CurrentTxRate5", FMT_WORDS, FW_COMMON},
	{0xfd85, "CurrentTxRate6", FMT_WORDS, FW_COMMON},
	{0xfd86, "OwnMACAddress", FMT_BYTES, FW_COMMON},
	{0xfd87, "PCFInfo", FMT_WORDS, FW_COMMON},
	{0xfd88, "HighestBasicRate", FMT_WORDS, FW_AGERE},
	{0xfd88, "ScanResults", FMT_WORDS, FW_INTERSIL},
	{0xfd89, "CurrentCountryInfo", FMT_WORDS, FW_AGERE},
	{0xfd89, "HostScanResults", FMT_WORDS, FW_INTERSIL},
	{0xfd8a, "CurrentSSNInfoElement", FMT_WORDS, FW_AGERE},
	{0xfd8a, "AuthenticationUsed", FMT_WORDS, FW_INTERSIL},
	{0xfd8b, "CurrentTKIPIVInfo", FMT_WORDS, FW_AGERE},
	{0xfd8b, "FAASwitchCtrl", FMT_WORDS, FW_INTERSIL},
	{0xfd8c, "AssociationRequestInfo", FMT_WORDS, FW_AGERE},
	{0xfd8d, "AssociationResponseInfo", FMT_WORDS, FW_AGERE},
	{0xfd8d, "AssociateFailure", FMT_WORDS, FW_INTERSIL},
	{0xfd8e, "CurrentLoad", FMT_WORDS, FW_AGERE},
	{0xfd90, "SecurityCapabilities", FMT_WORDS, FW_AGERE},
	{0xfdc0, "PHYType", FMT_WORDS, FW_COMMON},
	{0xfdc1, "CurrentChannel", FMT_WORDS, FW_COMMON},
	{0xfdc2, "CurrentPowerState", FMT_WORDS, FW_COMMON},
	{0xfdc3, "CCAMode", FMT_WORDS, FW_COMMON},
	{0xfdc6, "SupportedDataRates", FMT_BSTRING, FW_COMMON},
	{0xfdc7, "LFOStatus", FMT_WORDS, FW_INTERSIL},
	{0xfffe, "BuildSequence", FMT_WORDS, FW_COMMON},
	{0xffff, "FWID", FMT_FWSTRING, FW_COMMON},
};

#define ORINOCO_GETRID (SIOCIWFIRSTPRIV + 9)

static int orinoco_get_rid(int rid, void *buf, int *length)
{
	struct iwreq req;

	memset(&req, 0, sizeof(req));
	strncpy(req.ifr_name, iface, IFNAMSIZ);
	req.u.data.flags = rid;
	req.u.data.pointer = buf;
	req.u.data.length = MAX_RID_LEN;
	if (ioctl(sock, ORINOCO_GETRID, &req) < 0)
		return errno;

	*length = req.u.data.length;

	return 0;
}

#define HOSTAP_HDR_LEN 16
#define HOSTAP_GETRID_SUBCMD 9
#define HOSTAP_IOCTL (SIOCDEVPRIVATE + 14)

struct hostap_param {
	u32 cmd;
	u8 skip[8];
	u16 rid;
	u16 len;
	u8 data[0];
};

static int hostap_get_rid(int rid, void *buf, int *length)
{
	struct iwreq iwr;
	struct hostap_param *param = (struct hostap_param *)buf;

	param->cmd = HOSTAP_GETRID_SUBCMD;
	param->rid = rid;
	param->len = MAX_RID_LEN - HOSTAP_HDR_LEN;

	memset(&iwr, 0, sizeof(iwr));
	strncpy(iwr.ifr_name, iface, IFNAMSIZ);
	iwr.u.data.pointer = (caddr_t) param;
	iwr.u.data.length = MAX_RID_LEN;

	if (ioctl(sock, HOSTAP_IOCTL, &iwr) < 0)
		return errno;

	*length = param->len;
	memcpy(buf, param->data, param->len);

	return 0;
}

static int get_rid(int rid, void *buf, int *len)
{
	int err = 0;

	switch (driver_type) {
	case DRIVER_ORINOCO:
		err = orinoco_get_rid(rid, buf, len);
		break;
	case DRIVER_HOSTAP:
		err = hostap_get_rid(rid, buf, len);
		break;
	case DRIVER_UNKNOWN:
		err = EOPNOTSUPP;
		break;
	}

	return err;
}


#define CISTPL_VERS_1 0x15
#define CISTPL_MANFID 0x20
#define CISTPL_END 0xFF

static void print_cis(const u8 *cis, int len)
{
	int pos = 0;
	int manfid1 = -1, manfid2 = -1;
	int i, p, ilen;
	int ofs[4] = {0, 0, 0, 0};

	while (cis[pos] != CISTPL_END) {
		ilen = cis[pos + 1];

		if (pos + ilen >= len + 2)
			goto invalid;

		switch (cis[pos]) {
		case CISTPL_MANFID:
			if (ilen < 4)
				goto invalid;
			manfid1 = read_le16(cis + pos + 2);
			manfid2 = read_le16(cis + pos + 4);
			break;

		case CISTPL_VERS_1:
			if (ilen < 8)
				goto invalid;
			p = pos + 4;
			for (i = 0; i < 4; i++) {
				ofs[i] = p;
				while((p < pos + ilen) && cis[p])
					p++;
				while((p < pos + ilen) && !cis[p])
					p++;
				if (p >= pos + ilen)
					break;
			}
			break;
		}

		pos += ilen + 2;
	}

	if (ofs[0] == 0 || manfid1 == -1)
		goto invalid;

	printf("%04X:%04X", manfid1, manfid2);

	for (i = 0; i < 4; i++) {
		if (ofs[i])
			printf(" \"%s\"", &cis[ofs[i]]);
	}
	return;

 invalid:
	printf("invalid");
}


static void show_rid(int rid, char *name, display_t displaytype)
{
	u8 buf[MAX_RID_LEN];
	int err;

	int len;
	int j;

	memset(buf, 0, sizeof(buf));

	err = get_rid(rid, buf, &len);

	if (err) {
		if (err == ENODATA)
			return;

		fprintf(stderr, "Error for RID 0x%04x: %s.  "
			"Aborting.\n", rid, strerror(err));
		exit(1);
	}

	if (len <= 0)
		return;

	if (len > MAX_RID_LEN) {
		fprintf(stderr, "RID 0x%04x is too long: %d bytes.  "
			"Truncating to %d bytes\n", rid, len, MAX_RID_LEN);
		len = MAX_RID_LEN;
	}

	if (out_type != FMT_UNKNOWN)
		displaytype = out_type;
	else if (displaytype == FMT_UNKNOWN) {
		name = "unknown";
		displaytype = FMT_WORDS;
	}

	printf("%04x %3d %-25s ", rid, len, name);

	switch (displaytype) {
	case FMT_WORDS:
	case FMT_UNKNOWN:
		for (j = 0; j < len - 2; j += 2)
			printf("%04X-", read_le16(buf + j));
		printf("%04X", read_le16(buf + j));
		break;

	case FMT_BYTES:
		for (j = 0; j < len - 1; j++)
			printf("%02X:", buf[j]);
		printf("%02X", buf[j]);
		break;

	case FMT_STRING:
		len = min_t(u16, len - 2, read_le16(buf));
		printf("\"%.*s\"", len, buf + 2);
		break;

	case FMT_BSTRING:
		len = min_t(u16, len - 2, read_le16(buf));
		for (j = 2; j < len + 1; j++)
			printf("%02X,", buf[j]);
		printf("%02X", buf[j]);
		break;

	case FMT_XSTRING:
		printf("'%.*s'", len, buf);
		break;

	case FMT_CIS:
		print_cis(buf, len);
		break;

	case FMT_FWSTRING:
		printf("'%.14s' '%.*s'", buf, len - 14, buf + 14);
		break;
	}

	printf("\n");
}

/* Return position in record_table.  We assume that RIDs in record_table
 * are sorted.  */
static int search_rid(u16 rid)
{
	int i;

	for (i = 0; i < sizeof(record_table) / sizeof(record_table[0]); i++) {
		if (record_table[i].rid > rid)
			return -1;
		if (record_table[i].rid < rid)
			continue;
		if (record_table[i].fwtypes & fw_type)
			return i;
	}
	return -1;
}

static void show_known_rids(void)
{
	int i;

	for (i = 0; i < sizeof(record_table) / sizeof(record_table[0]); i++) {
		u16 rid = record_table[i].rid;
		char *name = record_table[i].name;
		int displaytype = record_table[i].displaytype;
		if (record_table[i].fwtypes & fw_type)
			show_rid(rid, name, displaytype);
	}
}

static void show_unknown_rids(void)
{
	int rid;

	for (rid = 0xfc00; rid <= 0xffff; rid++) {
		int i = search_rid(rid);

		if (i == -1)
			show_rid(rid, NULL, FMT_UNKNOWN);
	}
}

static void show_one_rid(int rid)
{
	char *name;
	int displaytype;

	int i = search_rid(rid);

	if (i == -1) {
		name = NULL;
		displaytype = FMT_UNKNOWN;
	} else {
		name = record_table[i].name;
		displaytype = record_table[i].displaytype;
	}
	show_rid(rid, name, displaytype);
}

static void show_all_rids(void)
{
	int rid;

	for (rid = 0xfc00; rid <= 0xffff; rid++)
		show_one_rid(rid);
}

#define SIOCETHTOOL 0x8946
#define ETHTOOL_GDRVINFO 0x00000003

struct ethtool_drvinfo {
	u32 cmd;
	char driver[32];
	char other_things[160];
};

static driver_t determine_driver_type(void)
{
	struct ifreq ifr;
	struct ethtool_drvinfo drvinfo;

	strncpy(ifr.ifr_ifrn.ifrn_name, iface, IFNAMSIZ);
	drvinfo.cmd = ETHTOOL_GDRVINFO;
	ifr.ifr_data = (caddr_t) & drvinfo;

	if (ioctl(sock, SIOCETHTOOL, &ifr) < 0) {
		fprintf(stderr, "ethtool ioctl failed: %s\n", strerror(errno));
		return DRIVER_UNKNOWN;
	}

	if (!strncmp(drvinfo.driver, "hostap", 6))
		return DRIVER_HOSTAP;
	if (!strncmp(drvinfo.driver, "orinoco", 7))
		return DRIVER_ORINOCO;

	fprintf(stderr, "Unknown driver: %s\n", drvinfo.driver);
	return DRIVER_UNKNOWN;
}

struct comp_id {
	u16 id, variant, major, minor;
} __attribute__ ((packed));

static int determine_firmware_type(void)
{
	u8 buf[MAX_RID_LEN];
	struct comp_id *nic_id;
	int len;
	int err;

	err = get_rid(HERMES_RID_NICID, buf, &len);

	if (err) {
		fprintf(stderr, "Error reading NICID: %s.  Aborting.\n",
			strerror(err));
		exit(1);
	}

	if (len != sizeof(struct comp_id)) {
		fprintf(stderr, "Unexpected NICID length: %d.  Aborting.\n",
			len);
		exit(1);
	}

	nic_id = (struct comp_id *)buf;

	if (nic_id->id < 0x8000)
		return FW_AGERE;
	else if (nic_id->id == 0x8000 && nic_id->major == 0)
		return FW_SYMBOL;
	else
		return FW_INTERSIL;
}

static void usage(void)
{
	printf("Usage: dump_recs [options] interface\n"
	       "Options:\n"
	       "  --all                Dump all RIDs\n"
	       "  --unknown            Dump unknown RIDs\n"
	       "  --rid RID            Dump this RID only\n"
	       "  --driver DRIVER      Use API of this driver.  Supported drivers:\n"
	       "                       orinoco, hostap\n"
	       "  --format FORMAT      Force output format.  Supported formats:\n"
	       "                       bytes, words, string, bstring, xstring,"
	       "                       fwstring, cis\n");
}

int main(int argc, char **argv)
{
	struct iwreq req;
	int err;
	char *rid_str = NULL;
	int rid = 0;
	char *type_str = NULL;
	char *drv_str = NULL;
	int unknown_flag = 0;
	int all_flag = 0;

	while (1) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"version", no_argument, 0, 'V'},
			{"all", no_argument, 0, 'a'},
			{"unknown", no_argument, 0, 'u'},
			{"rid", required_argument, 0, 'r'},
			{"format", required_argument, 0, 'f'},
			{"driver", required_argument, 0, 'd'},
			{0, 0, 0, 0}
		};

		c = getopt_long(argc, argv, "auhVr:f:d:", long_options,
				&option_index);

		if (c == -1)
			break;

		switch (c) {
		case 'V':
			printf("dump_recs version " PROGRAM_VERSION "\n");
			exit(0);
			break;

		case 'h':
			usage();
			exit(0);
			break;

		case 'a':
			all_flag = 1;
			break;

		case 'u':
			unknown_flag = 1;
			break;

		case 'r':
			rid_str = optarg;
			break;

		case 'f':
			type_str = optarg;
			break;

		case 'd':
			drv_str = optarg;
			break;

		default:
			usage();
			exit(1);
			break;
		}
	}

	if (optind == argc - 1)
		iface = argv[optind];
	else {
		usage();
		exit(1);
	}

	if (rid_str) {
		rid = strtol(rid_str, NULL, 16);
		if (rid < 0xfc00 || rid > 0xffff) {
			fprintf(stderr, "RID must be between 0xfc00 and "
				"0xffff\n");
			exit(1);
		}
	}

	if ((rid != 0) + all_flag + unknown_flag > 1) {
		fprintf(stderr, "Incompatible options\n");
		usage();
		exit(1);
	}

	out_type = FMT_UNKNOWN;
	if (type_str) {
		if (!strcasecmp(type_str, "bytes"))
			out_type = FMT_BYTES;
		else if (!strcasecmp(type_str, "words"))
			out_type = FMT_WORDS;
		else if (!strcasecmp(type_str, "string"))
			out_type = FMT_STRING;
		else if (!strcasecmp(type_str, "bstring"))
			out_type = FMT_BSTRING;
		else if (!strcasecmp(type_str, "xstring"))
			out_type = FMT_XSTRING;
		else if (!strcasecmp(type_str, "cis"))
			out_type = FMT_CIS;
		else if (!strcasecmp(type_str, "fwstring"))
			out_type = FMT_FWSTRING;
		else {
			fprintf(stderr, "Output type \"%s\" not recognized\n",
				type_str);
			usage();
			exit(1);
		}
		if (out_type != FMT_BYTES && out_type != FMT_WORDS && !rid) {
			fprintf(stderr, "Output type \"%s\" should not be "
				"used without RID\n", type_str);
			exit(1);
		}
	}

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("Cannot create socket:");
		exit(1);
	}

	if (drv_str) {
		if (!strcasecmp(drv_str, "orinoco"))
			driver_type = DRIVER_ORINOCO;
		else if (!strcasecmp(drv_str, "hostap"))
			driver_type = DRIVER_HOSTAP;
		else {
			fprintf(stderr, "Driver \"%s\" not recognized\n",
				drv_str);
			usage();
			exit(1);
		}
	}

	if (driver_type == DRIVER_UNKNOWN)
		driver_type = determine_driver_type();

	if (driver_type == DRIVER_UNKNOWN) {
		fprintf(stderr, "Cannot determine driver type for "
			"interface \"%s\"\n", iface);
		exit(1);
	}

	memset(&req, 0, sizeof(req));
	strncpy(req.ifr_name, iface, IFNAMSIZ);
	err = ioctl(sock, SIOCGIWNAME, &req);
	if (err) {
		fprintf(stderr, "Interface \"%s\" doesn't support "
			"wireless extensions\n", iface);
		exit(1);
	}

	fw_type = determine_firmware_type();
	printf("Firmware vendor: ");
	switch (fw_type) {
	case FW_AGERE:
		printf("Agere\n");
		break;
	case FW_SYMBOL:
		printf("Symbol\n");
		break;
	case FW_INTERSIL:
		printf("Intersil\n");
		break;
	}

	printf("RID  Len Name                      Data\n");

	if (rid)
		show_one_rid(rid);
	else if (all_flag)
		show_all_rids();
	else if (unknown_flag)
		show_unknown_rids();
	else
		show_known_rids();

	return 0;
}
